#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import unittest
from label_sleuth.data_access.data_access_api import get_document_uri


class TestDataAccessApi(unittest.TestCase):
    def test_get_document_uri_with_date_and_time(self):
        doc_uri = get_document_uri("imported-15_Jan_2023_17-51-28-683")
        self.assertEqual("imported-15_Jan_2023_17-51-28", doc_uri)

    def test_get_document_uri(self):
        doc_uri = get_document_uri("wiki_animals_2000_pages-Bobak marmot-22")
        self.assertEqual("wiki_animals_2000_pages-Bobak marmot", doc_uri)


if __name__ == "__main__":
    unittest.main()
